/*
 *  C implementation of the Aho-Corasic pattern matching algorithm. It's based
 *  on ScannerDaemon's Java version by Kurt Huwig and
 *  http://www-sr.informatik.uni-tuebingen.de/~buehler/AC/AC.html
 *  Thanks for Kurt Huwig for pointing me to this page.
 *
 *  Copyright (C) 2002 Tomasz Kojm <zolw@konarski.edu.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "others.h"
#include "shared.h"
#include "readdb.h"
#include "matcher.h"

void addpatt(struct node *root, struct patt *pattern)
{
	struct node *pos, *next;
	int i;

    if(pattern->length < MIN_LENGTH) {
	mprintf("!String too short.\n");
	return;
    }

    pos = root;

    for(i = 0; i < MIN_LENGTH; i++) {
	next = pos->next[((unsigned char) pattern->pattern[i]) & 0xff];

	if(next == NULL) {
	    next = (struct node *) mcalloc(1, sizeof(struct node));
	    pos->next[((unsigned char) pattern->pattern[i]) & 0xff] = next;
	}

	pos = next;
    }

    pos->islast = 1;

    pattern->next = pos->list;
    pos->list = pattern;
}

struct nodelist *bfsadd(struct nodelist *bfs, struct node *n)
{
	struct nodelist *new;

    new = (struct nodelist *) mcalloc(1, sizeof(struct nodelist));

    new->next = bfs;
    new->node = n;
    bfs = new;

    return bfs;
}

void failtrans(struct node *root)
{
	struct nodelist *bfs = NULL, *bfsnext = NULL, *pt, *prev;
	struct node *n, *q, *qs, *r;
	int i;

    root->fail = NULL; /* h(root) = fail */

    for(i = 0; i < NUM_CHILDS; i++) {
	/* for all childs of root */
	n = root->next[i];
	if(n) {
	    n->fail = root;
	    bfs = bfsadd(bfs, n);
	}
    }

    while(bfs) {
	bfsnext = NULL;

	pt = bfs;

	while(pt) {
	    q = pt->node;

	    for(i = 0; i < NUM_CHILDS; i++) {
		qs = q->next[i];

		if(qs) {
		    bfsnext = bfsadd(bfsnext, qs);
		    r = q->fail;
		    while(r && (r->next[i] == NULL))
			r = r->fail;

		    if(r == NULL)
			qs->fail = root->next[i];
		    else
			qs->fail = r->next[i];
		}
	    }

	    pt = pt->next;
	}

	/* free list */

	pt = bfs;
	while(pt) {
	    prev = pt;
	    pt = pt->next;
	    free(prev);
	}

	bfs = bfsnext;
    }
}

void fasttrie(struct node *n, struct node *root)
{
	int i;
	struct node *next, *q;


    for(i = 0; i < NUM_CHILDS; i++) {
	next = n->next[i];
	if(next) {
	    n->trans[i] = next;
	    if(!n->islast)
		fasttrie(next, root);
	} else {
	    q = n;
	    do {
		q = q->fail;
	    } while(q && (q->next[i] == NULL));
	    n->trans[i] = q ? q->next[i] : root;
	}
    }
}

/* return virus name or NULL if not found */
char *scanbuff(const char *buffer, int length, const struct node *root)
{
	struct node *current;
	struct patt *pt;
	int i, position;

    current = (struct node *) root;

    for(i = 0; i < length; i++)  {
	current = current->trans[(unsigned char) buffer[i] & 0xff];

	if(current->islast) {
	    position = i - MIN_LENGTH + 1;

	    pt = current->list;
	    while(pt) {
		if(findpos(buffer, position, length, pt)) {
		    return pt->virname;
		}

		pt = pt->next;
	    }

	}
    }

    return NULL;
}

int findpos(const char *buffer, int offset, int length, const struct patt *pattern)
{
	int bufferpos = offset + MIN_LENGTH;
	int postfixend = offset + length;
	int i;


    for(i = MIN_LENGTH; i < pattern->length; i++) {

	bufferpos %= length;

	if(bufferpos == postfixend)
	    return 0;

	if(pattern->pattern[i] != buffer[bufferpos])
	    return 0;

	bufferpos++;
    }

    return 1;
}

/* scan given descriptor and return virus name or NULL */
char *scandesc(int desc, const struct node *root)
{
	char *buffer, *buff, *endbl, *virname;
	int bytes, buffsize;


    /* prepare the buffer */
    buffsize = root->maxpatlen + BUFFSIZE;
    buffer = (char *) mcalloc(buffsize, sizeof(char));

    buff = buffer;
    buff += root->maxpatlen; /* pointer to read data block */
    endbl = buff + BUFFSIZE - root->maxpatlen; /* pointer to the last block
						* length of root->maxpatlen
						*/

    while((bytes = read(desc, buff, BUFFSIZE)) > 0) {

	claminfo.blocks += bytes / 16384; /* count 16kb blocks */

	if((virname = scanbuff(buffer, buffsize, root))) {
	    free(buffer);
	    return virname;
	}

	if(bytes == BUFFSIZE)
	    memmove(buffer, endbl, root->maxpatlen);

    }

    free(buffer);
    return NULL;
}

int checkfile(const char *filename, const struct node *root)
{
	int fd;
	char *virname = NULL;

    if((fd = open(filename, O_RDONLY)) == -1) {
	mprintf("@Can't open file %s\n", filename);
	return 54;
    }

    claminfo.files++;

    if((virname = scandesc(fd, root))) {
	mprintf("%s: %s FOUND\n", filename, virname);
	claminfo.ifiles++;
    } else if(!printinfected)
	mprintf("%s: OK\n", filename);

    close(fd);
    return virname ? 1 : 0;
}

int checkstdin(const struct node *root)
{
	char *virname = NULL;

    claminfo.files++;

    if((virname = scandesc(0, root))) {
	mprintf("stdin: %s FOUND\n", virname);
	claminfo.ifiles++;
    } else if(!printinfected)
	mprintf("stdin: OK\n");

    return virname ? 1 : 0;
}
